#ifndef i2c_sniff_h
/*
 * i2c sniffer by storepeter (C) 2022 Beerware
 * sends sampled data on rs232 @ 2Mbit for further analysis on linux
 *
 * https://en.wikipedia.org/wiki/I%C2%B2C
 * for timing i2c timing see https://rheingoldheavy.com/i2c-signals/
 *     ___    _   _   _   _   _   _   _   _   _     _   _   _   _   _   _   _   _   _   _____
 * SCL    \__/1\_/2\_/3\_/4\_/5\_/6\_/7\_/8\_/9\___/1\_/2\_/3\_/4\_/5\_/6\_/7\_/8\_/9\_/
 *     _     ___ ___ ___ ___ ___ ___ ___ ___       ___ ___ ___ ___ ___ ___ ___ ___ ___     __
 * SDA  \___X___X___X___X___X___X___X___X___X_____X___X___X___X___X___X___X___X___X   \___/
 *     Start <----- 7 bit addres ------>  rw  ack  <-------- 8 bit Data ---------> Nack  Stop
 * 
 * state                SCL     SDA     comment
 * (N) inactive         1       1       pulled up, free to claim arbitration
 * (S) Start            1       \_      controller pull SDA low (falling edge)
 * (i) idle             1       0       held low by controller
 * (0) bit=0            _/      0       SDA unchanged while SCL high (rising edge)
 * (0) bit=0            \_      0       controller pull SCL low (falling edge)
 * (1) bit=1            _/      1       SDA unchanged while SCL high (rising edge)
 * (1) bit=1            \_      1       controller pull SCL low (falling edge)
 * (P) Stop             1       _/      controller releases SDA  (rising edge)
 * (CS) clock stretch   0       X       target pulls SCL low until SDA set (if rd) (while it was low)
 * (Sr) Repeated Start  1       \_      Same as start but not stop condition since last start
 *
 *  7bit address: S adr7 R/W Ack data8 Ack data8 Nack Stop
 * S xxxxxxx r/w ack
 * 10bit address: S 11110 adr2 R/W Ack adr8 data8 Ack data8 Nack Stop
 * S 11110xx r/w ack xxxxxxxx ack
 *
 * SMBus is i2c @ 100kbit
 *
 * The Protocol send to linux is simple
 * State 2 bit + 4 bit data,
 *
 * START + 4 msb
 * DATA  + 4 msb
 * Ack   + 4 lsb
 * Nack  + 4 lsb
 * STOP  + 4bit time since last START, logarithmic
 * IDLE  + 4bit time since last STOP/IDLE, logarithmic, idle -> 2 Hz heartbeat
 */
#define START   0x20
#define DATA    0x30
#define ACK     0x40
#define NACK    0x50
#define STOP    0x60
#define IDLE    0x70
#ifdef RECORD_SAMPLING_TIME
#define SAMPLE_CHK      0x90
#endif
/* timing 400kHz, minimum xfer would be 20 bit i.e. 20kHz = 50 usec. is enough resolution
 * Timer prescaler options
 * 16 MHz  8_prescaler -> 500 nsec
 * 16 MHz  64_prescaler -> 4 usec
 * 16 MHz  256_prescaler -> 16 usec  This i what we will use
 * 16 MHz  1024_prescaler -> 64 usec
 * 0x0001 0x01  16 usec
 * 0x0002 0x02  32 usec
 * 0x0004 0x03  64 usec
 * 0x0008 0x04 128 usec
 * 0x0010 0x05 256 usec
 * 0x0020 0x06 512 usec
 * 0x0040 0x07   1 msec
 * 0x0080 0x08   2 msec
 * 0x0100 0x09   4 msec
 * 0x0200 0x0a   8 msec
 * 0x0400 0x0b  16 msec
 * 0x0800 0x0c  32 msec
 * 0x1000 0x0d  64 msec
 * 0x2000 0x0e 128 msec
 * 0x4000 0x0f 256 msec
 * 0x8000 0x10 512 msec
 *
 * 0xFFF8 0x80  1 sec   i.e 1Hz heartbeat when no trafic
 */
/* S adr7 R/W Ack  - data8 Ack - data8 Nack - Stop
 * 0x2? 0x4?       - 0x3? 0x4? - 0x3? 0x5?  - 0x6? 0x80 0x80..
 * logarithmic idle timer:
 * STOP=0x63 ->  32 usec -> 400kBit 20bit  50usec
 * STOP-0x65 -> 128 usec -> 100kBit 20bit 200usec
 */
//    prev    currrent
//    scl sda scl sda
//      0 0     0 0     -
//      0 0     0 1     normal setting of SDA
//      0 0     1 0     SDA bit=0  stable until scl_pulldown
//      0 0     1 1     UNEXPECTED SDA should not change when SCL changes
//      0 1     0 0     normal clearing of SDA
//      0 1     0 1     -
//      0 1     1 0     UNEXPECTED SDA should not change when SCL changes
//      0 1     1 1     SDA bit=1 stable until scl_pulldown
//      1 0     0 0     bit=0   scl_pulldown
//      1 0     0 1     bit=1   scl_pulldown
//      1 0     1 0     -
//      1 0     1 1     STOP condition
//      1 1     0 0     bit=0   scl_pulldown
//      1 1     0 1     bit=1   scl_pulldown
//      1 1     1 0     STAR conditionT
//      1 1     1 1     -
 
#endif  // i2c_sniff_h
